xpath = require('./xpath')
Util = require('./util')
$ = require('jquery')

Range = {}

# Public: Determines the type of Range of the provided object and returns
# a suitable Range instance.
#
# r - A range Object.
#
# Examples
#
#   selection = window.getSelection()
#   Range.sniff(selection.getRangeAt(0))
#   # => Returns a BrowserRange instance.
#
# Returns a Range object or false.
Range.sniff = (r) ->
  if r.commonAncestorContainer?
    new Range.BrowserRange(r)
  else if typeof r.start is "string"
    new Range.SerializedRange(r)
  else if r.start and typeof r.start is "object"
    new Range.NormalizedRange(r)
  else
    false

class Range.RangeError extends Error
  constructor: (@type, @message, @parent = null) ->
    super(@message)

# Public: Creates a wrapper around a range object obtained from a DOMSelection.
class Range.BrowserRange

  # Public: Creates an instance of BrowserRange.
  #
  # object - A range object obtained via DOMSelection#getRangeAt().
  #
  # Examples
  #
  #   selection = window.getSelection()
  #   range = new Range.BrowserRange(selection.getRangeAt(0))
  #
  # Returns an instance of BrowserRange.
  constructor: (obj) ->
    @commonAncestorContainer = obj.commonAncestorContainer
    @startContainer          = obj.startContainer
    @startOffset             = obj.startOffset
    @endContainer            = obj.endContainer
    @endOffset               = obj.endOffset

  # Public: normalize works around the fact that browsers don't generate
  # ranges/selections in a consistent manner. Some (Safari) will create
  # ranges that have (say) a textNode startContainer and elementNode
  # endContainer. Others (Firefox) seem to only ever generate
  # textNode/textNode or elementNode/elementNode pairs.
  #
  # Returns an instance of Range.NormalizedRange
  normalize: (root) ->
    if @tainted
      return false
    else
      @tainted = true

    r = {}
    this._normalizeStart(r)
    this._normalizeEnd(r)

    # Now let's start to slice & dice the text elements!
    nr = {}

    if r.startOffset > 0
      # Do we really have to cut?
      if r.start.nodeValue.length > r.startOffset
        # Yes. Cut.
        nr.start = r.start.splitText(r.startOffset)
      else
        # Avoid splitting off zero-length pieces.
        nr.start = r.start.nextSibling
    else
      nr.start = r.start

    # is the whole selection inside one text element ?
    if r.start is r.end
      if nr.start.nodeValue.length > (r.endOffset - r.startOffset)
        nr.start.splitText(r.endOffset - r.startOffset)
      nr.end = nr.start
    else # no, the end of the selection is in a separate text element
      # does the end need to be cut?
      if r.end.nodeValue.length > r.endOffset
        r.end.splitText(r.endOffset)
      nr.end = r.end

    # Make sure the common ancestor is an element node.
    nr.commonAncestor = @commonAncestorContainer
    while nr.commonAncestor.nodeType isnt Util.NodeTypes.ELEMENT_NODE
      nr.commonAncestor = nr.commonAncestor.parentNode

    new Range.NormalizedRange(nr)

  _normalizeStart: (r) ->
    # Look at the start
    if @startContainer.nodeType is Util.NodeTypes.ELEMENT_NODE
      # We are dealing with element nodes
      r.start = Util.getFirstTextNodeNotBefore(
        @startContainer.childNodes[@startOffset]
      )
      r.startOffset = 0
    else
      # We are dealing with simple text nodes
      r.start = @startContainer
      r.startOffset = @startOffset

  _normalizeEnd: (r) ->
    # Look at the end
    if @endContainer.nodeType is Util.NodeTypes.ELEMENT_NODE
      # Get specified node.
      node = @endContainer.childNodes[@endOffset]

      if node? # Does that node exist?
        # Look for a text node either at the immediate beginning of node
        n = node
        while n? and (n.nodeType isnt Util.NodeTypes.TEXT_NODE)
          n = n.firstChild
        if n? # Did we find a text node at the start of this element?
          r.end = n
          r.endOffset = 0

      unless r.end?
        # We need to find a text node in the previous sibling of the node at the
        # given offset, if one exists, or in the previous sibling of its
        # container.
        if @endOffset
          node = @endContainer.childNodes[@endOffset - 1]
        else
          node = @endContainer.previousSibling
        r.end = Util.getLastTextNodeUpTo node
        r.endOffset = r.end.nodeValue.length

    else # We are dealing with simple text nodes
      r.end = @endContainer
      r.endOffset = @endOffset

  # Public: Creates a range suitable for storage.
  #
  # root           - A root Element from which to anchor the serialisation.
  # ignoreSelector - A selector String of elements to ignore. For example
  #                  elements injected by the annotator.
  #
  # Returns an instance of SerializedRange.
  serialize: (root, ignoreSelector) ->
    this.normalize(root).serialize(root, ignoreSelector)

# Public: A normalised range is most commonly used throughout the annotator.
# its the result of a deserialised SerializedRange or a BrowserRange with
# out browser inconsistencies.
class Range.NormalizedRange

  # Public: Creates an instance of a NormalizedRange.
  #
  # This is usually created by calling the .normalize() method on one of the
  # other Range classes rather than manually.
  #
  # obj - An Object literal. Should have the following properties.
  #       commonAncestor: A Element that encompasses both the start and end
  #                       nodes
  #       start:          The first TextNode in the range.
  #       end             The last TextNode in the range.
  #
  # Returns an instance of NormalizedRange.
  constructor: (obj) ->
    @commonAncestor = obj.commonAncestor
    @start          = obj.start
    @end            = obj.end

  # Public: For API consistency.
  #
  # Returns itself.
  normalize: (root) ->
    this

  # Public: Limits the nodes within the NormalizedRange to those contained
  # withing the bounds parameter. It returns an updated range with all
  # properties updated. NOTE: Method returns null if all nodes fall outside
  # of the bounds.
  #
  # bounds - An Element to limit the range to.
  #
  # Returns updated self or null.
  limit: (bounds) ->
    nodes = $.grep this.textNodes(), (node) ->
      node.parentNode == bounds or $.contains(bounds, node.parentNode)

    return null unless nodes.length

    @start = nodes[0]
    @end   = nodes[nodes.length - 1]

    startParents = $(@start).parents()
    for parent in $(@end).parents()
      if startParents.index(parent) != -1
        @commonAncestor = parent
        break
    this

  # Convert this range into an object consisting of two pairs of (xpath,
  # character offset), which can be easily stored in a database.
  #
  # root - The root Element relative to which XPaths should be calculated
  # ignoreSelector - A selector String of elements to ignore. For example
  #                  elements injected by the annotator.
  #
  # Returns an instance of SerializedRange.
  serialize: (root, ignoreSelector) ->

    serialization = (node, isEnd) ->
      if ignoreSelector
        origParent = $(node).parents(":not(#{ignoreSelector})").eq(0)
      else
        origParent = $(node).parent()

      path = xpath.fromNode(origParent, root)[0]
      textNodes = Util.getTextNodes(origParent)

      # Calculate real offset as the combined length of all the
      # preceding textNode siblings. We include the length of the
      # node if it's the end node.
      nodes = textNodes.slice(0, textNodes.index(node))
      offset = 0
      for n in nodes
        offset += n.nodeValue.length

      if isEnd then [path, offset + node.nodeValue.length] else [path, offset]

    start = serialization(@start)
    end   = serialization(@end, true)

    new Range.SerializedRange({
      # XPath strings
      start: start[0]
      end: end[0]
      # Character offsets (integer)
      startOffset: start[1]
      endOffset: end[1]
    })

  # Public: Creates a concatenated String of the contents of all the text nodes
  # within the range.
  #
  # Returns a String.
  text: ->
    (for node in this.textNodes()
      node.nodeValue
    ).join ''

  # Public: Fetches only the text nodes within th range.
  #
  # Returns an Array of TextNode instances.
  textNodes: ->
    textNodes = Util.getTextNodes($(this.commonAncestor))
    [start, end] = [textNodes.index(this.start), textNodes.index(this.end)]
    # Return the textNodes that fall between the start and end indexes.
    $.makeArray textNodes[start..end]

# Public: A range suitable for storing in local storage or serializing to JSON.
class Range.SerializedRange

  # Public: Creates a SerializedRange
  #
  # obj - The stored object. It should have the following properties.
  #       start:       An xpath to the Element containing the first TextNode
  #                    relative to the root Element.
  #       startOffset: The offset to the start of the selection from obj.start.
  #       end:         An xpath to the Element containing the last TextNode
  #                    relative to the root Element.
  #       startOffset: The offset to the end of the selection from obj.end.
  #
  # Returns an instance of SerializedRange
  constructor: (obj) ->
    @start       = obj.start
    @startOffset = obj.startOffset
    @end         = obj.end
    @endOffset   = obj.endOffset

  # Public: Creates a NormalizedRange.
  #
  # root - The root Element from which the XPaths were generated.
  #
  # Returns a NormalizedRange instance.
  normalize: (root) ->
    range = {}

    for p in ['start', 'end']
      try
        node = xpath.toNode(this[p], root)
      catch e
        throw new Range.RangeError(
          p,
          "Error while finding #{p} node: #{this[p]}: " + e,
          e
        )

      if not node
        throw new Range.RangeError(p, "Couldn't find #{p} node: #{this[p]}")

      # Unfortunately, we *can't* guarantee only one textNode per
      # elementNode, so we have to walk along the element's textNodes until
      # the combined length of the textNodes to that point exceeds or
      # matches the value of the offset.
      length = 0
      targetOffset = this[p + 'Offset']

      # Range excludes its endpoint because it describes the boundary position.
      # Target the string index of the last character inside the range.
      if p is 'end' then targetOffset -= 1

      for tn in Util.getTextNodes($(node))
        if (length + tn.nodeValue.length > targetOffset)
          range[p + 'Container'] = tn
          range[p + 'Offset'] = this[p + 'Offset'] - length
          break
        else
          length += tn.nodeValue.length

      # If we fall off the end of the for loop without having set
      # 'startOffset'/'endOffset', the element has shorter content than when
      # we annotated, so throw an error:
      if not range[p + 'Offset']?
        throw new Range.RangeError(
          "#{p}offset",
          "Couldn't find offset #{this[p + 'Offset']} in element #{this[p]}"
        )

    # Here's an elegant next step...
    #
    #   range.commonAncestorContainer = $(range.startContainer)
    #     .parents()
    #     .has(range.endContainer)[0]
    #
    # ...but unfortunately Node.contains() is broken in Safari 5.1.5 (7534.55.3)
    # and presumably other earlier versions of WebKit. In particular, in a
    # document like
    #
    #   <p>Hello</p>
    #
    # the code
    #
    #   p = document.getElementsByTagName('p')[0]
    #   p.contains(p.firstChild)
    #
    # returns `false`. Yay.
    #
    # So instead, we step through the parents from the bottom up and use
    # Node.compareDocumentPosition() to decide when to set the
    # commonAncestorContainer and bail out.

    contains =
      if document.compareDocumentPosition?
        # Everyone else
        (a, b) -> a.compareDocumentPosition(b) &
          Node.DOCUMENT_POSITION_CONTAINED_BY

      else
        # Newer IE
        (a, b) -> a.contains(b)

    $(range.startContainer).parents().each ->
      if range.endContainer.nodeType == Util.NodeTypes.TEXT_NODE
        endContainer = range.endContainer.parentNode
      else
        endContainer = range.endContainer

      if contains(this, endContainer)
        range.commonAncestorContainer = this
        return false

    new Range.BrowserRange(range).normalize(root)

  # Public: Creates a range suitable for storage.
  #
  # root           - A root Element from which to anchor the serialisation.
  # ignoreSelector - A selector String of elements to ignore. For example
  #                  elements injected by the annotator.
  #
  # Returns an instance of SerializedRange.
  serialize: (root, ignoreSelector) ->
    this.normalize(root).serialize(root, ignoreSelector)

  # Public: Returns the range as an Object literal.
  toObject: ->
    {
      start: @start
      startOffset: @startOffset
      end: @end
      endOffset: @endOffset
    }


# Export Range object.
module.exports = Range
